/************************************************************************
 *   IRC - Internet Relay Chat, ircd/s_serv.c (formerly ircd/s_msg.c)
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers. 
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include <string.h>
#include "h.h"

static char buf[BUFSIZE];


/*
** m_functions execute protocol messages on this server:
**
**	cptr	is always NON-NULL, pointing to a *LOCAL* client
**		structure (with an open socket connected!). This
**		identifies the physical socket where the message
**		originated (or which caused the m_function to be
**		executed--some m_functions may call others...).
**
**	sptr	is the source of the message, defined by the
**		prefix part of the message if present. If not
**		or prefix not found, then sptr==cptr.
**
**		(!IsServer(cptr)) => (cptr == sptr), because
**		prefixes are taken *only* from servers...
**
**		(IsServer(cptr))
**			(sptr == cptr) => the message didn't
**			have the prefix.
**
**			(sptr != cptr && IsServer(sptr) means
**			the prefix specified servername. (?)
**
**			(sptr != cptr && !IsServer(sptr) means
**			that message originated from a remote
**			user (not local).
**
**		combining
**
**		(!IsServer(sptr)) means that, sptr can safely
**		taken as defining the target structure of the
**		message in this server.
**	*Always* true (if 'parse' and others are working correct):
**
**	1)	sptr->from == cptr  (note: cptr->from == cptr)
**
**	2)	MyConnect(sptr) <=> sptr == cptr (e.g. sptr
**		*cannot* be a local connection, unless it's
**		actually cptr!). [MyConnect(x) should probably
**		be defined as (x == x->from) --msa ]
**
**	parc	number of variable parameter strings (if zero,
**		parv is allowed to be NULL)
**
**	parv	a NULL terminated list of parameter pointers,
**
**			parv[0], sender (prefix string), if not present
**				this points to an empty string.
**			parv[1]...parv[parc-1]
**				pointers to additional parameters
**			parv[parc] == NULL, *always*
**
**		note:	it is guaranteed that parv[0]..parv[parc-1] are all
**			non-NULL pointers.
*/
/* DL 2000-06-28 A-lines (code inspired by m_gline) */
int
m_aline (cptr, sptr, parc, parv)
     aClient *cptr, *sptr;
     int parc;
     char *parv[];
{
  aAline *aaline, *a2aline = NULL;

  if (IsServer (cptr))
    {
      if (find_conf_host (cptr->confs, sptr->name, CONF_ULINE))
	{
	  if (parc < 2)
	    {
	      sendto_one (sptr, err_str (ERR_NEEDMOREPARAMS), me.name,
			  parv[0], "ALINE");
	      return 0;
	    }

	  collapse (parv[1]);

	  sendto_serv_butone (cptr, ":%s ALINE :%s", parv[0], parv[1]);

	  if (*parv[1] == '-')
	    {
	      parv[1]++;
	      for (aaline = aline, a2aline = NULL; aaline;
		   aaline = aaline->next)
		{
		  if (!mycmp (aaline->pattern, parv[1]))
		    break;
		  a2aline = aaline;
		}
	      if (aaline)
		{
		  sendto_realops ("%s removing ALINE for %s", parv[0],
				  parv[1]);
		  /* notify opers */
		  if (a2aline)
		    {
		      a2aline->next = aaline->next;
		    }
		  else
		    {
		      aline = aaline->next;
		    }
		  MyFree (aaline->pattern);
		  MyFree (aaline);
		  write_alines ();
		}
	      else
		{
		  /* No A-line, do nothing */
		}
	    }
	  else
	    {
	      if (*parv[1] == '+')
		{
		  parv[1]++;
		}
	      for (aaline = aline, a2aline = NULL; aaline;
		   aaline = aaline->next)
		{
		  if (!mycmp (aaline->pattern, parv[1]))
		    break;
		  a2aline = aaline;
		}
	      if (aaline)
		{
		  /* A-line exists, do nothing */
		  ;
		}
	      else
		{
		  sendto_realops ("%s adding ALINE for %s", parv[0], parv[1]);
		  /* inform ops */
		  if (!aline)
		    {
		      aline = aaline =
			(struct Aline *) MyMalloc (sizeof (aAline));
		    }
		  else
		    {
		      a2aline->next = aaline =
			(struct Aline *) MyMalloc (sizeof (aAline));
		    }
		  DupString (aaline->pattern, parv[1]);
		  aaline->next = NULL;
		  write_alines ();
		}
	    }
	}
      else
	return 0;		/* server not permitted to add a ALINE anyway */
    }
  else
    {
      /* A user... just give them the list */
      if (!IsAnOper (sptr))
	{
	  sendto_one (sptr, err_str (ERR_NOPRIVILEGES), me.name, parv[0]);
	  return 0;
	}

      for (aaline = aline; aaline; aaline = aaline->next)
	{
	  /* display the ALINE */
	  sendto_one (cptr, rpl_str (RPL_ALIST), me.name, parv[0],
		      aaline->pattern);
	}
      sendto_one (cptr, rpl_str (RPL_ENDOFALIST), me.name, parv[0]);

    }

  return 0;
}
