/************************************************************************
 *   IRC - Internet Relay Chat, ircd/s_serv.c (formerly ircd/s_msg.c)
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers. 
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include <string.h>
#include "h.h"

static char buf[BUFSIZE];


/*
** m_functions execute protocol messages on this server:
**
**	cptr	is always NON-NULL, pointing to a *LOCAL* client
**		structure (with an open socket connected!). This
**		identifies the physical socket where the message
**		originated (or which caused the m_function to be
**		executed--some m_functions may call others...).
**
**	sptr	is the source of the message, defined by the
**		prefix part of the message if present. If not
**		or prefix not found, then sptr==cptr.
**
**		(!IsServer(cptr)) => (cptr == sptr), because
**		prefixes are taken *only* from servers...
**
**		(IsServer(cptr))
**			(sptr == cptr) => the message didn't
**			have the prefix.
**
**			(sptr != cptr && IsServer(sptr) means
**			the prefix specified servername. (?)
**
**			(sptr != cptr && !IsServer(sptr) means
**			that message originated from a remote
**			user (not local).
**
**		combining
**
**		(!IsServer(sptr)) means that, sptr can safely
**		taken as defining the target structure of the
**		message in this server.
**	*Always* true (if 'parse' and others are working correct):
**
**	1)	sptr->from == cptr  (note: cptr->from == cptr)
**
**	2)	MyConnect(sptr) <=> sptr == cptr (e.g. sptr
**		*cannot* be a local connection, unless it's
**		actually cptr!). [MyConnect(x) should probably
**		be defined as (x == x->from) --msa ]
**
**	parc	number of variable parameter strings (if zero,
**		parv is allowed to be NULL)
**
**	parv	a NULL terminated list of parameter pointers,
**
**			parv[0], sender (prefix string), if not present
**				this points to an empty string.
**			parv[1]...parv[parc-1]
**				pointers to additional parameters
**			parv[parc] == NULL, *always*
**
**		note:	it is guaranteed that parv[0]..parv[parc-1] are all
**			non-NULL pointers.
*/

 /*
    ** m_server
    **  parv[0] = sender prefix
    **  parv[1] = servername
    **  parv[2] = hopcount
    ** P09: parv[3] = start timestamp
    **      parv[4] = link timestamp
    **      parv[5] = major protocol version: P09
    **      parv[parc-1] = serverinfo
  */
static int
a_kills_b_too (a, b)
     register aClient *a, *b;
{
  for (; b != a && b != &me; b = b->serv->up)
    return (a == b ? 1 : 0);
}

int
m_server (cptr, sptr, parc, parv)
     aClient *cptr, *sptr;
     int parc;
     char *parv[];
{
  Reg1 char *ch;
  Reg2 int i;
  Reg3 Link *lp;
  char info[REALLEN + 1], *inpath, *host, *s;
  aClient *acptr, *bcptr, *LHcptr;
  aConfItem *aconf, *bconf, *cconf, *lhconf;
  int hop, ret, active_lh_line = 0;
  u_short prot;
  time_t start_timestamp, timestamp = 0, recv_time, ghost = 0;

  recv_time = TStime ();
  info[0] = '\0';
  inpath = get_client_name (cptr, TRUE);
  if (parc < 2 || *parv[1] == '\0')
    {
      sendto_one (cptr, "ERROR :No servername");
      return exit_client (cptr, cptr, &me, "No servername");
    }
  /* Since .U5 a hopcount is required --Run */
  if (parc < 3 || (hop = atoi (parv[2])) == 0 ||
      (IsServer (cptr) && parc < ((Protocol (cptr) > 4) ? 7 : 4)))
    {
      sendto_one (sptr, err_str (ERR_NEEDMOREPARAMS),
		  me.name, parv[0], "SERVER");
      return exit_client (cptr, cptr, &me, "Need more parameters");
    }
  host = parv[1];
  /* Detect Protocol(cptr) */
  if (parc > 6)
    {
      if (strlen (parv[5]) != 3 || (parv[5][0] != 'P' && parv[5][0] != 'J'))
	return exit_client_msg (cptr, sptr, &me,
				"Bogus protocol (%s)", parv[5]);
      prot = atoi (parv[5] + 1);
      start_timestamp = atoi (parv[3]);
      timestamp = atoi (parv[4]);
      Debug ((DEBUG_INFO, "Got SERVER %s with timestamp [%s] age %lu (%lu)",
	      host, parv[4], start_timestamp, me.serv->timestamp));
      if (prot > 4 &&
	  (timestamp < 780000000 || hop == 1 && start_timestamp < 780000000))
	{
	  return exit_client_msg (cptr, sptr, &me,
				  "Bogus timestamps (%s %s)", parv[3],
				  parv[4]);
	}
    }
  else if (!IsServer (cptr))	/* New local server ? */
    /* Protocol 4 didn't support protocol versions yet */
    prot = 4;
  else				/* Remote server */
    prot = 0;			/* Set to 'unknown' */
  if (parc > 3)
    (void) strncpy (info, parv[parc - 1], REALLEN);
  if (prot && (prot < atoi (MINOR_PROTOCOL)))
    {
      sendto_ops ("Got incompatible protocol version (%s) from %s",
		  parv[5], get_client_name (cptr, TRUE));
      return exit_new_server (cptr, sptr, host, prot, timestamp,
			      "Incompatible protocol: %s", parv[5]);
    }
  /*
     ** Check for "FRENCH " infection ;-) (actually this should
     ** be replaced with routine to check the hostname syntax in
     ** general). [ This check is still needed, even after the parse
     ** is fixed, because someone can send "SERVER :foo bar " ].
     ** Also, changed to check other "difficult" characters, now
     ** that parse lets all through... --msa
   */
  if (strlen (host) > HOSTLEN)
    host[HOSTLEN] = '\0';
  for (ch = host; *ch; ch++)
    if (*ch <= ' ' || *ch > '~')
      break;
  if (*ch || !index (host, '.'))
    {
      sendto_realops ("Bogus server name (%s) from %s", host,
		      inet_ntoa (cptr->ip.s_addr));
      return exit_client_msg (cptr, cptr, &me,
			      "Bogus server name (%s)", host);
    }

  if (IsPerson (cptr))
    {
      /*
         ** A local link that has been identified as a USER
         ** tries something fishy... ;-)
       */
      sendto_one (cptr, err_str (ERR_ALREADYREGISTRED), me.name, parv[0]);
      sendto_one (cptr, ":%s NOTICE %s :You realise that you suck?",
		  me.name, cptr->name);
      sendto_realops ("User %s trying to become a server %s",
		      get_client_name (cptr, TRUE), host);
      sendto_serv_butone (&me,
			  ":%s GLOBOPS :User %s trying to become a server %s",
			  me.name, get_client_name (cptr, TRUE), host);
      return exit_client_msg (cptr, cptr, &me, "I suck, really I do");
    }

  if (IsServer (cptr))
    {
      /*
         ** A local server introduces a new server behind this link.
         ** Check if this is allowed according L:, H: and Q: lines.
       */
      if (info[0] == '\0')
	return exit_client_msg (cptr, cptr, &me,
				"No server info specified for %s", host);

      /*
         ** See if the newly found server is behind a guaranteed
         ** leaf (L-line). If so, close the link.
       */
      if ((lhconf = find_conf_host (cptr->confs, host, CONF_LEAF)) &&
	  (!lhconf->port || (hop > lhconf->port)))
	{
	  /*
	     ** L: lines normally come in pairs, here we try to
	     ** make sure that the oldest link is squitted, not
	     ** both.
	   */
	  active_lh_line = 1;
	  if (timestamp <= cptr->serv->timestamp)
	    LHcptr = NULL;	/* Kill incoming server */
	  else
	    LHcptr = cptr;	/* Squit ourselfs */
	}
#if 0
      else if (!(lhconf = find_conf_host (cptr->confs, host, CONF_HUB)) ||
	       (lhconf->port && (hop > lhconf->port)))
	{
	  aClient *ac3ptr;
	  active_lh_line = 2;
	  /* Look for net junction causing this: */
	  LHcptr = (Protocol (sptr) > 4) ? NULL :	/* incoming server */
	    cptr;		/* Or local if prot==4 */
	  if (Protocol (sptr) > 4 && *parv[5] != 'J')
	    for (ac3ptr = sptr; ac3ptr != &me; ac3ptr = ac3ptr->serv->up)
	      if (IsJunction (ac3ptr))
		{
		  LHcptr = ac3ptr;
		  break;
		}
	}
#endif
    }

  if (IsUnknown (cptr) || IsHandshake (cptr))
    {
      char *encr;

      /*
         ** A local link that is still in undefined state wants
         ** to be a SERVER. Check if this is allowed and change
         ** status accordingly...
       */
      /*
         ** If there is more then one server on the same machine
         ** that we try to connect to, it could be that the /CONNECT
         ** <mask> caused this connect to be put at the wrong place
         ** in the hashtable.
         ** --Run
       */
      if (IsHandshake (cptr) &&
	  hash_nick_name (cptr->name) != hash_nick_name (host))
	{
	  del_from_client_hash_table (cptr->name, cptr);
	  add_to_client_hash_table (host, cptr);
	}
      strncpyzt (cptr->name, host, sizeof (cptr->name));
      strncpyzt (cptr->info, info[0] ? info : me.name, sizeof (cptr->info));
      cptr->hopcount = hop;

      /* check connection rules */
      for (cconf = conf; cconf; cconf = cconf->next)
	if ((cconf->status == CONF_CRULEALL) &&
	    (matches (cconf->host, host) == 0))
	  if (crule_eval (cconf->passwd))
	    {
	      ircstp->is_ref++;
	      sendto_ops ("Refused connection from %s.",
			  get_client_host (cptr));
	      return exit_client (cptr, cptr, &me,
				  "Disallowed by connection rule");
	    }

      if (check_server_init (cptr))
	{
	  ircstp->is_ref++;
	  sendto_realops ("Received unauthorized connection from %s.",
			  get_client_host (cptr));
	  return exit_client (cptr, cptr, &me, "No C/N conf lines");
	}

      host = cptr->name;

      current_load_data.conn_count++;
      update_load ();

      if (!(aconf = find_conf (cptr->confs, host, CONF_NOCONNECT_SERVER)))
	{
	  ircstp->is_ref++;
	  sendto_ops ("Access denied. No N line for server %s", inpath);
	  return exit_client_msg (cptr, cptr, &me,
				  "Access denied. No N line for server %s",
				  inpath);
	}
      if (!(bconf = find_conf (cptr->confs, host, CONF_CONNECT_SERVER)))
	{
	  ircstp->is_ref++;
	  sendto_ops ("Only N (no C) field for server %s", inpath);
	  return exit_client_msg (cptr, cptr, &me,
				  "Only N (no C) field for server %s",
				  inpath);
	}

#ifdef CRYPT_LINK_PASSWORD
      /* passwd may be NULL. Head it off at the pass... */
      if (*cptr->passwd)
	{
	  extern char *crypt ();

	  encr = crypt (cptr->passwd, aconf->passwd);
	}
      else
	encr = "";
#else
      encr = cptr->passwd;
#endif /* CRYPT_LINK_PASSWORD */
      if (*aconf->passwd && !StrEq (aconf->passwd, encr))
	{
	  ircstp->is_ref++;
	  sendto_ops ("Access denied (passwd mismatch) %s", inpath);
	  return exit_client_msg (cptr, cptr, &me,
				  "No Access (passwd mismatch) %s", inpath);
	}
      bzero (cptr->passwd, sizeof (cptr->passwd));

#ifndef	HUB
      for (i = 0; i <= highest_fd; i++)
	if (local[i] && IsServer (local[i]))
	  {
	    active_lh_line = 3;
	    LHcptr = cptr;
	    break;
	  }
#endif
      if (!IsUnknown (cptr))
	{
	  s = (char *) index (aconf->host, '@');
	  *s = '\0';		/* should never be NULL */
	  Debug ((DEBUG_INFO, "Check Usernames [%s]vs[%s]",
		  aconf->host, cptr->username));
	  if (matches (aconf->host, cptr->username))
	    {
	      *s = '@';
	      ircstp->is_ref++;
	      sendto_ops ("Username mismatch [%s]v[%s] : %s",
			  aconf->host, cptr->username,
			  get_client_name (cptr, TRUE));
	      return exit_client (cptr, cptr, &me, "Bad Username");
	    }
	  *s = '@';
	}
    }


  /*
     ** We want to find IsConnecting() and IsHandshake() too,
     ** use find_client(). The second try is for hostmasking, although
     ** I doubt if that is still fully supported at all --Run.
   */
  while ((acptr = find_client (host, NULL)) ||
	 (acptr = find_name (host, NULL)))
    {
      /*
         ** This link is trying feed me a server that I already have
         ** access through another path
         **
         ** Do not allow U:Line to do this.
         ** Do not allow servers that are juped.
         ** Do not allow servers that have older link timestamps
         **   then this try.
         **
         ** If my ircd.conf sucks, I can try to connect to myself:
       */
      if (acptr == &me)
	return exit_client_msg (cptr, cptr, &me,
				"nick collision with me (%s)", host);
      /*
         ** Kill our try, if we had one.
       */
      if (IsConnecting (acptr))
	{
	  if (!active_lh_line && exit_client (cptr, acptr, &me,
					      "Just connected via another link")
	      == CPTR_KILLED)
	    return CPTR_KILLED;
	  /*
	     **  We can have only ONE 'IsConnecting', 'IsHandshake' or
	     ** 'IsServer', because new 'IsConnecting's are refused to
	     ** the same server if we already had it.
	   */
	  break;
	}
      /*
         ** Avoid other nick collisions...
         ** This is a doubtfull test though, what else would it be
         ** when it has a server.name ?
       */
      else if (!IsServer (acptr) && !IsHandshake (acptr))
	return exit_client_msg (cptr, cptr, &me,
				"Nickname %s already exists!", host);
      /*
         ** Our new server might be a juped server,
         ** or someone trying abuse a second U:Lines:
       */
      else if (IsServer (acptr) &&
	       (strncasecmp (acptr->info, "JUPE", 4) == 0 ||
		find_conf_host (cptr->confs, sptr->name, CONF_ULINE)))
	{
	  if (!IsServer (sptr))
	    return exit_client (cptr, sptr, &me, acptr->info);
	  sendto_one (cptr,
		      ":%s WALLOPS :Received :%s SERVER %s from %s !?!",
		      me.name, parv[0], parv[1], cptr->name);
	  return exit_new_server (cptr, sptr, host, prot, timestamp,
				  "%s", acptr->info);
	}
      /*
         ** Of course we find the handshake this link was before :)
       */
      else if (IsHandshake (acptr) && acptr == cptr)
	break;
      /*
         ** This is the old protocol, can be removed when all upgraded
         ** to u2.9.
       */
      if (prot < 9 || (IsServer (acptr) && Protocol (acptr) < 9))
	return exit_client_msg (cptr, cptr, &me,
				"server %s already exists (protocol 4 failure)",
				host);
      /*
         ** Here we have a server nick collision...
         ** We don't want to kill the link that was last /connected,
         ** but we neither want to kill a good (old) link.
         ** Therefor we kill the second youngest link.
       */
      else
	{
	  aClient *c2ptr = NULL, *c3ptr = acptr;
	  aClient *ac2ptr, *ac3ptr;

	  /* Search youngest link: */
	  for (ac3ptr = acptr; ac3ptr != &me; ac3ptr = ac3ptr->serv->up)
	    if (ac3ptr->serv->timestamp > c3ptr->serv->timestamp)
	      c3ptr = ac3ptr;
	  if (IsServer (sptr))
	    {
	      for (ac3ptr = sptr; ac3ptr != &me; ac3ptr = ac3ptr->serv->up)
		if (ac3ptr->serv->timestamp > c3ptr->serv->timestamp)
		  c3ptr = ac3ptr;
	    }
	  if (timestamp > c3ptr->serv->timestamp)
	    {
	      c3ptr = NULL;
	      c2ptr = acptr;	/* Make sure they differ */
	    }
	  /* Search second youngest link: */
	  for (ac2ptr = acptr; ac2ptr != &me; ac2ptr = ac2ptr->serv->up)
	    if (ac2ptr != c3ptr &&
		ac2ptr->serv->timestamp >
		(c2ptr ? c2ptr->serv->timestamp : timestamp))
	      c2ptr = ac2ptr;
	  if (IsServer (sptr))
	    {
	      for (ac2ptr = sptr; ac2ptr != &me; ac2ptr = ac2ptr->serv->up)
		if (ac2ptr != c3ptr &&
		    ac2ptr->serv->timestamp >
		    (c2ptr ? c2ptr->serv->timestamp : timestamp))
		  c2ptr = ac2ptr;
	    }
	  if (c3ptr && timestamp >
	      (c2ptr ? c2ptr->serv->timestamp : timestamp))
	    c2ptr = NULL;
	  /* If timestamps are equal, decide which link to break
	     ** by name.
	   */
	  if ((c2ptr ? c2ptr->serv->timestamp : timestamp) ==
	      (c3ptr ? c3ptr->serv->timestamp : timestamp))
	    {
	      char *n2, *n2up;
	      char *n3, *n3up;
	      if (c2ptr)
		{
		  n2 = c2ptr->name;
		  n2up = MyConnect (c2ptr) ? me.name : c2ptr->serv->up->name;
		}
	      else
		{
		  n2 = host;
		  n2up = IsServer (sptr) ? sptr->name : me.name;
		}
	      if (c3ptr)
		{
		  n3 = c3ptr->name;
		  n3up = MyConnect (c3ptr) ? me.name : c3ptr->serv->up->name;
		}
	      else
		{
		  n3 = host;
		  n3up = IsServer (sptr) ? sptr->name : me.name;
		}
	      if (strcmp (n2, n2up) > 0)
		n2 = n2up;
	      if (strcmp (n3, n3up) > 0)
		n3 = n3up;
	      if (strcmp (n3, n2) > 0)
		{
		  ac2ptr = c2ptr;
		  c2ptr = c3ptr;
		  c3ptr = ac2ptr;
		}
	    }
	  /* Now squit the second youngest link: */
	  if (!c2ptr)
	    return exit_new_server (cptr, sptr, host, prot, timestamp,
				    "server %s already exists and is %ld seconds younger.",
				    host,
				    (long) acptr->serv->timestamp -
				    (long) timestamp);
	  else if (c2ptr->from == cptr || IsServer (sptr))
	    {
	      if (active_lh_line)
		{
		  /*
		     ** If the L: or H: line also gets rid of this link,
		     ** we sent just one squit.
		   */
		  if (LHcptr && a_kills_b_too (LHcptr, c2ptr))
		    break;
		  /*
		     ** If breaking the loop here solves the L: or H:
		     ** line problem, we don't squit that.
		   */
		  if (c2ptr->from == cptr ||
		      (LHcptr && a_kills_b_too (c2ptr, LHcptr)))
		    active_lh_line = 0;
		  else
		    {
		      /*
		         ** If we still have a L: or H: line problem,
		         ** we prefer to squit the new server, solving
		         ** loop and L:/H: line problem with only one squit.
		       */
		      LHcptr = NULL;
		      break;
		    }
		}
	      /*
	         ** If the new server was introduced by a server that caused a
	         ** Ghost less then 20 seconds ago, this is probably also
	         ** a Ghost... (20 seconds is more then enough because all
	         ** SERVER messages are at the beginning of a net.burst). --Run
	       */
	      if (now - cptr->serv->ghost < 20)
		{
		  if (exit_client (cptr, acptr, &me, "Ghost loop") ==
		      CPTR_KILLED)
		    return CPTR_KILLED;
		}
	      else if (exit_client_msg (cptr, c2ptr, &me,
					"Loop <-- %s (new link is %d seconds younger)",
					host,
					(c3ptr ? (long) c3ptr->serv->
					 timestamp : timestamp) -
					(long) c2ptr->serv->timestamp) ==
		       CPTR_KILLED)
		return CPTR_KILLED;
	    }
	  else
	    {
	      if (active_lh_line)
		{
		  if (LHcptr && a_kills_b_too (LHcptr, acptr))
		    break;
		  if (acptr->from == cptr ||
		      (LHcptr && a_kills_b_too (acptr, LHcptr)))
		    active_lh_line = 0;
		  else
		    {
		      LHcptr = NULL;
		      break;
		    }
		}
	      /*
	         ** We can't believe it is a lagged server message
	         ** when it directly connects to us...
	         ** kill the older link at the ghost, rather then
	         ** at the second youngest link, assuming it isn't
	         ** a REAL loop.
	       */
	      ghost = now;	/* Mark that it caused a ghost */
	      if (exit_client (cptr, acptr, &me, "Ghost") == CPTR_KILLED)
		return CPTR_KILLED;
	      break;
	    }
	  /*
	     ** Did we kill the incoming server off already ?
	   */
	  if (c2ptr->from == cptr)
	    return 0;
	}
    }

  if (active_lh_line)
    {
      if (LHcptr == NULL)
	return exit_new_server (cptr, sptr, host, prot, timestamp,
				(active_lh_line == 2) ?
				"Non-Hub link %s <- %s(%s)" :
				"Leaf-only link %s <- %s(%s)",
				get_client_name (cptr, TRUE), host,
				lhconf ? (lhconf->host ? lhconf->
					  host : "*") : "!");
      else
	{
	  register int killed = a_kills_b_too (LHcptr, sptr);
	  if (active_lh_line < 3)
	    {
	      if (exit_client_msg (cptr, LHcptr, &me,
				   (active_lh_line == 2) ?
				   "Non-Hub link %s <- %s(%s)" :
				   "Leaf-only link %s <- %s(%s)",
				   get_client_name (cptr, TRUE), host,
				   lhconf ? (lhconf->host ? lhconf->
					     host : "*") : "!") ==
		  CPTR_KILLED)
		return CPTR_KILLED;
	    }
	  else
	    {
	      ircstp->is_ref++;
	      if (exit_client (cptr, LHcptr, &me, "I'm a leaf") ==
		  CPTR_KILLED)
		return CPTR_KILLED;
	    }
	  /*
	     ** Did we kill the incoming server off already ?
	   */
	  if (killed)
	    return 0;
	}
    }

  if (IsServer (cptr))
    {
      /*
         ** Server is informing about a new server behind
         ** this link. Create REMOTE server structure,
         ** add it to list and propagate word to my other
         ** server links...
       */

      acptr = make_client (cptr);
      (void) make_server (acptr);
      acptr->serv->prot = prot;
      acptr->serv->timestamp = timestamp;
      acptr->hopcount = hop;
      strncpyzt (acptr->name, host, sizeof (acptr->name));
      strncpyzt (acptr->info, info, sizeof (acptr->info));
      acptr->serv->up = sptr;
      acptr->serv->updown = add_dlink (&sptr->serv->down, acptr);
      SetServer (acptr);
      acptr->flags |= FLAGS_TS8;
      add_client_to_list (acptr);
      (void) add_to_client_hash_table (acptr->name, acptr);
      if (Protocol (cptr) > 4 && *parv[5] == 'J')
	{
	  sendto_ops ("Net junction: %s %s", sptr->name, acptr->name);
	  SetJunction (acptr);
	}

      /* We need to set FLAGS_ULINE to the U:Lined Server - Aka */

      if (IsULine (sptr) || (find_uline (cptr->confs, acptr->name)))
	{
	  acptr->flags |= FLAGS_ULINE;
	  sendto_ops ("%s U:lined server merged with %s", acptr->name,
		      cptr->name);
	}

      /*
         ** Old sendto_serv_but_one() call removed because we now
         ** need to send different names to different servers
         ** (domain name matching)
       */
      for (i = 0; i <= highest_fd; i++)
	{
	  if (!(bcptr = local[i]) || !IsServer (bcptr) ||
	      bcptr == cptr || IsMe (bcptr))
	    continue;
	  if (!(cconf = bcptr->serv->nline))
	    {
	      sendto_ops ("Lost N-line for %s on %s. Closing",
			  get_client_name (cptr, TRUE), host);
	      return exit_client (cptr, cptr, &me, "Lost N line");
	    }
	  if (matches (my_name_for_link (me.name, cconf), acptr->name) == 0)
	    continue;
	  if (Protocol (bcptr) > 4)
	    {
	      if (Protocol (cptr) > 4)
		sendto_one (bcptr, ":%s %s %s %d 0 %s %s :%s",
			    parv[0],
			    (Protocol (bcptr) >= 12) ? "S" : "SERVER",
			    acptr->name, hop + 1,
			    parv[4], parv[5], acptr->info);
	      else
		sendto_one (bcptr, ":%s %s %s %d 0 0 P0%u :%s",
			    parv[0],
			    (Protocol (bcptr) >= 12) ? "S" : "SERVER",
			    acptr->name, hop + 1,
			    Protocol (acptr), acptr->info);
	    }
	  else
	    sendto_one (bcptr, ":%s SERVER %s %d :%s",
			parv[0], acptr->name, hop + 1, acptr->info);
	}
      return 0;
    }

  if (IsUnknown (cptr) || IsHandshake (cptr))
    {
      (void) make_server (cptr);
      cptr->serv->timestamp = timestamp;
      cptr->serv->prot = prot;
      cptr->serv->ghost = ghost;

      ret = m_server_estab (cptr, aconf, bconf);
    }
  else
    ret = 0;

  if (IsServer (cptr) && prot > 4 && start_timestamp > 780000000 &&
      start_timestamp < me.serv->timestamp)
    {
#ifndef RELIABLE_CLOCK
#ifdef TESTNET
      sendto_ops ("Debug: my start time: %lu ; others start time: %lu",
		  me.serv->timestamp, start_timestamp);
      sendto_ops ("Debug: receive time: %lu ; received timestamp: %lu"
		  " ; difference %ld", recv_time, timestamp,
		  timestamp - recv_time);
#endif
      me.serv->timestamp += timestamp - recv_time;
      TSoffset += timestamp - recv_time;
#else
      if (abs (timestamp - recv_time) > 30)
	{
	  sendto_ops ("Connected to a net with a timestamp-clock"
		      " difference of %ld seconds! Used SETTIME to correct"
		      " this.", timestamp - recv_time);
	  sendto_one (cptr, ":%s SETTIME %lu :%s", me.name,
		      TStime (), me.name);
	}
#endif
    }

  return ret;
}

/*
 * m_server_estab, may only be called after a SERVER was received from cptr,
 * and thus make_server was called, and serv->prot set. --Run
 */
int
m_server_estab (cptr, aconf, bconf)
     Reg1 aClient *cptr;
     Reg2 aConfItem *aconf, *bconf;
{
  Reg3 aClient *acptr;
  Reg4 Link *lp;
  char *inpath, *host, *s;
  int split, i;

  split = (mycmp (cptr->name, cptr->sockhost) &&
	   strncasecmp (cptr->info, "JUPE", 4));
  inpath = get_client_name (cptr, TRUE);
  host = cptr->name;

  if (IsUnknown (cptr))
    {
      if (bconf->passwd[0])
	sendto_one (cptr, "PASS :%s", bconf->passwd);
      /*
         ** Pass my info to the new server
       */
      if (Protocol (cptr) == 4)
	sendto_one (cptr, "SERVER %s 1 :%s",
		    my_name_for_link (me.name, aconf),
		    (me.info[0]) ? (me.info) : "IRCers United");
      else
	sendto_one (cptr, "SERVER %s 1 %lu %lu J%s :%s",
		    my_name_for_link (me.name, aconf),
		    me.serv->timestamp, cptr->serv->timestamp,
		    MAJOR_PROTOCOL,
		    (me.info[0]) ? (me.info) : "IRCers United");
    }

  det_confs_butmask (cptr,
		     CONF_LEAF | CONF_HUB | CONF_NOCONNECT_SERVER |
		     CONF_ULINE);

  if (!IsHandshake (cptr))
    (void) add_to_client_hash_table (cptr->name, cptr);
  SetServer (cptr);
  cptr->flags |= FLAGS_TS8;
  nextping = now;

  /* We're setting FLAGS_ULINE to a U:Lined server and making pretty lil server notices */
  if ((find_uline (cptr->confs, cptr->name)))
    cptr->flags |= FLAGS_ULINE;

  if (cptr->serv->user && (acptr = find_client (cptr->serv->by, NULL))
      && acptr->user == cptr->serv->user)
    sendto_one (acptr, ":%s NOTICE %s :Link with %s established (%s).",
		me.name, acptr->name, inpath,
		IsULine (cptr) ? "ULined" : "Normal");
  else
    acptr = NULL;
  sendto_lops_butone (acptr, "Link with %s established (%s).", inpath,
		      IsULine (cptr) ? "ULined" : "Normal");
/* Hello */
  cptr->serv->up = &me;
  cptr->serv->updown = add_dlink (&me.serv->down, cptr);
  cptr->serv->nline = aconf;
  sendto_ops ("Net junction: %s %s", me.name, cptr->name);
  SetJunction (cptr);
  /*
     ** Old sendto_serv_but_one() call removed because we now
     ** need to send different names to different servers
     ** (domain name matching) Send new server to other servers.
   */
  for (i = 0; i <= highest_fd; i++)
    {
      if (!(acptr = local[i]) || !IsServer (acptr) ||
	  acptr == cptr || IsMe (acptr))
	continue;
      if ((aconf = acptr->serv->nline) &&
	  !matches (my_name_for_link (me.name, aconf), cptr->name))
	continue;
      if (split && Protocol (acptr) < 9)
	{
	  sendto_one (acptr, ":%s SERVER %s 2 :[%s] %s", me.name,
		      cptr->name, cptr->sockhost, cptr->info);
	}
      else if (split)
	{
	  sendto_one (acptr, ":%s %s %s 2 0 %lu %s%u :[%s] %s",
		      me.name, (Protocol (acptr) >= 12) ? "S" : "SERVER",
		      cptr->name, cptr->serv->timestamp,
		      (Protocol (cptr) > 9) ? "J" : "J0", Protocol (cptr),
		      cptr->sockhost, cptr->info);
	}
      else if (!split && Protocol (acptr) < 9)
	{
	  sendto_one (acptr, ":%s SERVER %s 2 :%s",
		      me.name, cptr->name, cptr->info);
	}
      else
	{
	  sendto_one (acptr, ":%s %s %s 2 0 %lu %s%u :%s",
		      me.name, (Protocol (acptr) >= 12) ? "S" : "SERVER",
		      cptr->name, cptr->serv->timestamp,
		      (Protocol (cptr) > 9) ? "J" : "J0", Protocol (cptr),
		      cptr->info);
	}
    }

  /*
     ** Pass on my client information to the new server
     **
     ** First, pass only servers (idea is that if the link gets
     ** cancelled beacause the server was already there,
     ** there are no NICK's to be cancelled...). Of course,
     ** if cancellation occurs, all this info is sent anyway,
     ** and I guess the link dies when a read is attempted...? --msa
     **
     ** Note: Link cancellation to occur at this point means
     ** that at least two servers from my fragment are building
     ** up connection this other fragment at the same time, it's
     ** a race condition, not the normal way of operation...
     **
   */

  aconf = cptr->serv->nline;
  for (acptr = &me; acptr; acptr = acptr->prev)
    {
      /* acptr->from == acptr for acptr == cptr */
      if (acptr->from == cptr)
	continue;
      if (IsServer (acptr))
	{
	  if (matches (my_name_for_link (me.name, aconf), acptr->name) == 0)
	    continue;
	  split = (MyConnect (acptr) &&
		   mycmp (acptr->name, acptr->sockhost) &&
		   strncasecmp (acptr->info, "JUPE", 4));
	  if (split && Protocol (cptr) < 9)
	    {
	      sendto_one (cptr, ":%s SERVER %s %d :[%s] %s",
			  acptr->serv->up->name, acptr->name,
			  acptr->hopcount + 1, acptr->sockhost, acptr->info);
	    }
	  else if (split)
	    {
	      sendto_one (cptr,
			  ":%s %s %s %d 0 %lu %s%u :[%s] %s",
			  acptr->serv->up->name,
			  (Protocol (cptr) >= 12) ? "S" : "SERVER",
			  acptr->name,
			  acptr->hopcount + 1, acptr->serv->timestamp,
			  (Protocol (acptr) > 9) ? "P" : "P0",
			  Protocol (acptr), acptr->sockhost, acptr->info);
	    }
	  else if (!split && Protocol (cptr) < 9)
	    {
	      sendto_one (cptr, ":%s SERVER %s %d :%s",
			  acptr->serv->up->name, acptr->name,
			  acptr->hopcount + 1, acptr->info);
	    }
	  else
	    {
	      sendto_one (cptr,
			  ":%s %s %s %d 0 %lu %s%u :%s",
			  acptr->serv->up->name,
			  (Protocol (cptr) >= 12) ? "S" : "SERVER",
			  acptr->name,
			  acptr->hopcount + 1, acptr->serv->timestamp,
			  (Protocol (acptr) > 9) ? "P" : "P0",
			  Protocol (acptr), acptr->info);
	    }
	}
    }

  for (acptr = &me; acptr; acptr = acptr->prev)
    {
      /* acptr->from == acptr for acptr == cptr */
      if (acptr->from == cptr)
	continue;
      if (IsPerson (acptr))
	{
	  /*
	     ** IsPerson(x) is true only when IsClient(x) is true.
	     ** These are only true when *BOTH* NICK and USER have
	     ** been received. -avalon
	     ** Or only NICK in new format. --Run
	   */
	  if (Protocol (cptr) <= 9)
	    sendto_one (cptr, ":%s NICK %s %d %d %s %s %s :%s",
			acptr->user->server->name,
			acptr->name, acptr->hopcount + 1, acptr->lastnick,
			acptr->user->username, acptr->user->host,
			acptr->user->server->name, acptr->info);
	  else if (Protocol (cptr) <= 11)
	    sendto_one (cptr, ":%s N %s %d %d %s %s %s :%s",
			acptr->user->server->name,
			acptr->name, acptr->hopcount + 1, acptr->lastnick,
			acptr->user->username, acptr->user->host,
			acptr->user->server->name, acptr->info);
	  else if (Protocol (cptr) <= 12)
	    sendto_one (cptr, ":%s N %s %d %d %s %s %s %s %s %d :%s",
			acptr->user->server->name,
			acptr->name, acptr->hopcount + 1, acptr->lastnick,
			acptr->user->username, acptr->user->host,
			acptr->user->virthost, acptr->user->server->name,
			send_umode (NULL, acptr, 0, SEND_UMODES, buf),
			acptr->user->helpstatus, acptr->info);
	  else if (Protocol (cptr) >= 13)	/* make this <= when it changes */
	    sendto_one (cptr, ":%s N %s %d %d %s %s %s %s %s %d %lu :%s",
			acptr->user->server->name,
			acptr->name, acptr->hopcount + 1, acptr->lastnick,
			acptr->user->username, acptr->user->host,
			acptr->user->virthost, acptr->user->server->name,
			send_umode (NULL, acptr, 0, SEND_UMODES, buf),
			acptr->user->helpstatus, acptr->ip.s_addr,
			acptr->info);

	  if (Protocol (cptr) <= 11)
	    send_umode (cptr, acptr, 0, SEND_UMODES, buf);

	  send_user_joins (cptr, acptr);
	}
      else if (IsService (acptr))
	{
	  sendto_one (cptr, "NICK %s :%d", acptr->name, acptr->hopcount + 1);
	  sendto_one (cptr, ":%s SERVICE * * :%s", acptr->name, acptr->info);
	}
    }
  /*
     ** Last, pass all channels plus statuses
   */
  {
    Reg1 aChannel *chptr;
    for (chptr = channel; chptr; chptr = chptr->nextch)
      {
	send_channel_modes (cptr, chptr);
	if (chptr->topic_time)
	  {
	    if (Protocol (cptr) <= 10)
	      sendto_one (cptr, ":%s TOPIC %s %s %lu :%s",
			  me.name, chptr->chname, chptr->topic_nick,
			  chptr->topic_time, chptr->topic);
	    if (Protocol (cptr) >= 11)
	      sendto_one (cptr, ":%s T %s %s %lu :%s",
			  me.name, chptr->chname, chptr->topic_nick,
			  chptr->topic_time, chptr->topic);
	  }
      }

  }
/* CHANGE THIS WHEN SERVICES GOES TO P09 */
  if (Protocol (cptr) >= 4)
    sendto_one (cptr, ":%s END_OF_BURST", me.name);
  sendto_ops ("Completed sending Burst Data. %lu bytes sent to %s.",
	      cptr->bsend, cptr->name);
  return 0;

}
