/************************************************************************
 *   IRC - Internet Relay Chat, ircd/s_serv.c (formerly ircd/s_msg.c)
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers. 
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include <string.h>
#include "h.h"

static char buf[BUFSIZE];

/*
** m_functions execute protocol messages on this server:
**
**	cptr	is always NON-NULL, pointing to a *LOCAL* client
**		structure (with an open socket connected!). This
**		identifies the physical socket where the message
**		originated (or which caused the m_function to be
**		executed--some m_functions may call others...).
**
**	sptr	is the source of the message, defined by the
**		prefix part of the message if present. If not
**		or prefix not found, then sptr==cptr.
**
**		(!IsServer(cptr)) => (cptr == sptr), because
**		prefixes are taken *only* from servers...
**
**		(IsServer(cptr))
**			(sptr == cptr) => the message didn't
**			have the prefix.
**
**			(sptr != cptr && IsServer(sptr) means
**			the prefix specified servername. (?)
**
**			(sptr != cptr && !IsServer(sptr) means
**			that message originated from a remote
**			user (not local).
**
**		combining
**
**		(!IsServer(sptr)) means that, sptr can safely
**		taken as defining the target structure of the
**		message in this server.
**	*Always* true (if 'parse' and others are working correct):
**
**	1)	sptr->from == cptr  (note: cptr->from == cptr)
**
**	2)	MyConnect(sptr) <=> sptr == cptr (e.g. sptr
**		*cannot* be a local connection, unless it's
**		actually cptr!). [MyConnect(x) should probably
**		be defined as (x == x->from) --msa ]
**
**	parc	number of variable parameter strings (if zero,
**		parv is allowed to be NULL)
**
**	parv	a NULL terminated list of parameter pointers,
**
**			parv[0], sender (prefix string), if not present
**				this points to an empty string.
**			parv[1]...parv[parc-1]
**				pointers to additional parameters
**			parv[parc] == NULL, *always*
**
**		note:	it is guaranteed that parv[0]..parv[parc-1] are all
**			non-NULL pointers.
*/
   /*
      ** m_squit
      **        parv[0] = sender prefix
      **        parv[1] = server name
      **      parv[2] = timestamp
      **        parv[parc-1] = comment
    */
int
m_squit (cptr, sptr, parc, parv)
     aClient *cptr, *sptr;
     int parc;
     char *parv[];
{
  Reg1 aConfItem *aconf;
  char *server;
  Reg2 aClient *acptr;
  char *comment = (parc > ((!IsServer (cptr) || Protocol (cptr) < 9) ?
			   2 : 3) && !BadPtr (parv[parc - 1])) ?
    parv[parc - 1] : cptr->name;

  if (!IsPrivileged (sptr))
    {
      sendto_one (sptr, err_str (ERR_NOPRIVILEGES), me.name, parv[0]);
      return 0;
    }

  if (parc > (IsServer (cptr) ? ((Protocol (cptr) < 9) ? 1 : 2) : 1))
    {
      server = parv[1];
      /*
         ** To accomodate host masking, a squit for a masked server
         ** name is expanded if the incoming mask is the same as
         ** the server name for that link to the name of link.
       */
      if ((*server == '*') && IsServer (cptr) &&
	  (aconf = cptr->serv->nline) &&
	  !mycmp (server, my_name_for_link (me.name, aconf)))
	{
	  server = cptr->name;
	  acptr = cptr;
	}
      else
	{
	  /*
	     ** The following allows wild cards in SQUIT. Only usefull
	     ** when the command is issued by an oper.
	   */
	  for (acptr = client; (acptr = next_client (acptr, server));
	       acptr = acptr->next)
	    if (IsServer (acptr) || IsMe (acptr))
	      break;

	  if (acptr)
	    {
	      if (IsMe (acptr))
		{
		  if (IsServer (cptr))
		    {
		      acptr = cptr;
		      server = cptr->sockhost;
		    }
		  else
		    acptr = NULL;
		}
	      else
		{
		  /*
		     ** Look for a matching server that is closer,
		     ** that way we won't accidently squit two close
		     ** servers like davis.* and davis-r.* when typing
		     ** /SQUIT davis*
		   */
		  aClient *acptr2;
		  for (acptr2 = acptr->serv->up; acptr2 != &me;
		       acptr2 = acptr2->serv->up)
		    if (!match (server, acptr2->name))
		      acptr = acptr2;
		}
	    }
	}
      /* If atoi(parv[2]) == 0 we must indeed squit !
         ** It wil be our neighbour.
       */
      if (acptr && IsServer (sptr) && Protocol (sptr) > 4 &&
	  atoi (parv[2]) && atoi (parv[2]) != acptr->serv->timestamp)
	{
	  Debug ((DEBUG_NOTICE, "Ignoring SQUIT with wrong timestamp"));
	  return 0;
	}
    }
  else
    {
      sendto_one (cptr, err_str (ERR_NEEDMOREPARAMS),
		  me.name, parv[0], "SQUIT");
      if (IsServer (cptr))
	{
	  /*
	     ** This is actually protocol error. But, well, closing
	     ** the link is very proper answer to that...
	   */
	  server = cptr->sockhost;
	  acptr = cptr;
	}
      else
	return 0;
    }
  if (!acptr)
    {
      if (IsPerson (sptr))
	sendto_one (sptr, err_str (ERR_NOSUCHSERVER),
		    me.name, parv[0], server);
      return 0;
    }
  if (IsLocOp (sptr) && !MyConnect (acptr))
    {
      sendto_one (sptr, err_str (ERR_NOPRIVILEGES), me.name, parv[0]);
      return 0;
    }
  /*
     **  Notify all opers, if my local link is remotely squitted
   */
  if (MyConnect (acptr) && !IsAnOper (cptr))
    {
      sendto_ops ("Received SQUIT %s from %s (%s)",
		  acptr->name, get_client_name (sptr, FALSE), comment);
#if defined(USE_SYSLOG) && defined(SYSLOG_SQUIT)
      syslog (LOG_DEBUG, "SQUIT From %s : %s (%s)", parv[0], server, comment);
#endif
    }
  else if (MyConnect (acptr))
    {
      sendto_ops ("Received SQUIT %s from %s (%s)",
		  acptr->name, get_client_name (sptr, FALSE), comment);
    }

  return exit_client (cptr, acptr, sptr, comment);
}

int
exit_new_server (cptr, sptr, host, prot, timestamp, fmt, p1, p2, p3)
     aClient *cptr, *sptr;
     u_short prot;
     time_t timestamp;
     char *fmt, *p1, *p2, *p3, *host;
{
  char buf[128];
  if (!IsServer (sptr))
    {
      if (prot < 9)
	{
	  strcpy (buf, ":%s SQUIT %s :");
	  strcat (buf, fmt);
	  sendto_one (cptr, buf, me.name, host, p1, p2, p3);
	}
      return exit_client_msg (cptr, cptr, &me, fmt, p1, p2, p3);
    }
  if (Protocol (cptr) < 9)
    {
      strcpy (buf, ":%s SQUIT %s :");
      strcat (buf, fmt);
      sendto_one (cptr, buf, me.name, host, p1, p2, p3);
    }
  else
    {
      strcpy (buf, ":%s SQUIT %s %lu :");
      strcat (buf, fmt);
      sendto_one (cptr, buf, me.name, host, timestamp, p1, p2, p3);
    }
  return 0;
}

static int
a_kills_b_too (a, b)
     register aClient *a, *b;
{
  for (; b != a && b != &me; b = b->serv->up)
    return (a == b ? 1 : 0);
}
